/*
  (c) 2005-2007 Philipp Klaus Krause pkk@spth.de

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.
   
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
   
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <iostream>
#include <fstream>
#include <ios>

#include <stdint.h>
#include <stdlib.h>
#include <ctype.h>

// Modify notes by halftones.
signed int global_modifiers[7] = {0, 0, 0, 0, 0, 0, 0};
signed int local_modifiers[4][7] = {{0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}};
bool local_modifier_enabled[4][7] = {{false, false, false, false, false, false, false}, {false, false, false, false, false, false, false}, {false, false, false, false, false, false, false}, {false, false, false, false, false, false, false}};
const signed int halftones[7] = {0, 2, 4, 5, 7, 9, 11};
const unsigned char C = 0;
const unsigned char D = 1;
const unsigned char E = 2;
const unsigned char F = 3;
const unsigned char G = 4;
const unsigned char A = 5;
const unsigned char B = 6;
const unsigned char Z = 23;

const unsigned int LENGTH_SCALE = 16;

std::ofstream ofile;

unsigned int default_note;

unsigned int length;
unsigned int tlength;

unsigned int legato;

static struct
{
	unsigned char base;
	signed int oktave;
	signed int local_modifier;
	bool local_modifier_enabled;
	char num[16];
	bool denom_current;
	unsigned int counter;
	char denom[16];
	unsigned char loudness;
	bool local_legato;
	bool staccato;
} note;

void erase(void)
{
	note.base = 0;
	note.local_modifier = 0;
	note.local_modifier_enabled = false;
	note.denom_current = false;
	memset(note.num, 0, 16);
	memset(note.denom, 0, 16);
	note.counter = 0;
	note.local_legato = false;
	note.staccato =false;
}

void translate(void)
{
	if(!note.base) return;
	length++;
	signed int transformed;
	switch(note.base)
	{
	case 'C':
		note.base = C;
		break;
	case 'D':
		note.base = D;
		break;
	case 'E':
		note.base = E;
		break;
	case 'F':
		note.base = F;
		break;
	case 'G':
		note.base = G;
		break;
	case 'A':
		note.base = A;
		break;
	case 'B':
		note.base = B;
		break;
	case 'Z':
		note.base = Z;
		break;
	default:
		erase();
		return;
	}
	signed char modifier;
	if(note.base != Z)
	{
		if(note.local_modifier_enabled)
		{
			modifier = note.local_modifier;
			local_modifiers[note.oktave][note.base] = modifier;
			local_modifier_enabled[note.oktave][note.base] = true;
		}
		else if(local_modifier_enabled[note.oktave][note.base])
		{
			modifier = local_modifiers[note.oktave][note.base];
		}
		else
			modifier = global_modifiers[note.base];
		transformed = halftones[note.base];
		transformed += modifier;
		if(transformed < 0)
		{
			transformed += 12;
			note.oktave--;
		}
		if(transformed >= 12)
		{
			transformed -= 12;
			note.oktave++;
		}
	}
	else	// Pause
	{
		note.oktave = 0;
		transformed = 0xf;
	}

	unsigned int length = default_note;
	unsigned int rel_length;
	if(note.num[0])
		length *= atoi(note.num);
	if(note.denom[0])
		length /= atoi(note.denom);
	else if(note.denom_current)
		length /= 2;
	uint8_t tone_byte;
	note.loudness = 2;	// Default to mf.
	if(note.local_legato || legato)
		rel_length = 0;	// Full length.
	else if(note.staccato)
		rel_length = 1;	// 25%.
	else
		rel_length = 3;	// default length 75%.
	uint8_t length_byte = (length << 4) | (rel_length << 2) | note.loudness;
	uint16_t entry;
	tone_byte = (note.oktave << 4) | transformed;
	entry = (uint16_t(tone_byte) << 8) | uint16_t(length_byte);
	ofile << " " << entry << ",";
	tlength += length;
	erase();
}

void usage(void)
{
	std::cout << "abc2cvm 0.6\n";
	std::cout << "Usage: abc2cvm [--verbose][-v] filename\n";
}

void unimplemented_key(void)
{
	std::cerr << "Unimplemented key.\n";
	std::cerr << "Supported keys:\nC, G, D, A, E, B, F, Bb, Eb, Ab, Db and Gb (major scale)\nAm, Dm, Gm, Cm, Fm, Bm, and Em (minor scale).\nOnly violin keys are implemented.\n";
}

int main(int argc, char **argv)
{
	std::string ifilename, ofilename;
	bool verbose = false;

	legato = 0;
	length = 0;
	tlength = 0;

	int i;
	if(argc == 1)
	{
		std::cerr << "Too few arguments.\n";
		usage();
		return(-1);
	}
	for(i = 1; i < argc - 1; i++)
	{
		std::string stra = argv[i];
		if(stra == "--verbose" || stra == "-v")
		{
			verbose = true;
		}
		else
		{
			usage();
			return(-1);
		}
	}

	ifilename = argv[argc - 1];
	if(verbose) std::cout << "Using " << ifilename << " as input file.\n";
	ofilename = ifilename + ".music.c";
	if(verbose) std::cout << "Using " << ofilename << " as output file.\n";

	std::ifstream ifile; 
	ifile.open(ifilename.c_str(), std::ios::in);
	ofile.open(ofilename.c_str(), std::ios::out | std::ios::trunc);

	char a;
	while((a = ifile.get()) != EOF)
	{
		if(a == 'K')
		{
			ifile.get();	// Get ':'
			while((a = ifile.get()) == ' ');	// Skip whitespace
			char b = ifile.get();
			if(b != '\n' && b != ' ' && b != '%' && b!= 'b' && b!= 'm')
			{
				unimplemented_key();
				return(-1);
			}
			
			if(b != 'b')	// Major scales
			{
				switch(a)
				{
				case 'B':
					global_modifiers[A] = 1;
				case 'E':
					global_modifiers[D] = 1;
				case 'A':
					global_modifiers[G] = 1;
				case 'D':
					global_modifiers[C] = 1;
				case 'G':
					global_modifiers[F] = 1;
				case 'C':
					break;
				case 'F':
					global_modifiers[B] = -1;
					break;
				default:
					unimplemented_key();
					return(-1);
				}
			}
			else if(b != 'm')	// Major scales
			{
				switch(a)
				{
				case 'G':
					global_modifiers[C] = -1;
				case 'D':
					global_modifiers[G] = -1;
				case 'A':
					global_modifiers[D] = -1;
				case 'E':
					global_modifiers[A] = -1;
				case 'B':
					global_modifiers[E] = -1;
					global_modifiers[B] = -1;
					break;
				default:
					unimplemented_key();
					return(-1);
				}
			}
			else	// Minor scales
			{
				switch(a)
				{
				case 'F':
					global_modifiers[D] = -1;
				case 'C':
					global_modifiers[A] = -1;
				case 'G':
					global_modifiers[E] = -1;
				case 'D':
					global_modifiers[B] = -1;
				case 'A':
					break;
				case 'B':
					global_modifiers[F] = 1;
				case 'E':
					global_modifiers[C] = 1;
					break;
				default:
					unimplemented_key();
					return(-1);
				}
			}
			// Ignore clef.
			if(b != '\n')
				while(ifile.get() != '\n');
			break;
		}
		else if(a == 'M')
		{
			std::cerr << "Setting metrum not implemented yet (M header entry).\n";
			while(ifile.get() != '\n');
		}
		else if(a == 'L')
		{
			if(ifile.get() != ':') continue;
			char temp[16] = "               ";
			i = 0;
			while((a = ifile.get()) != '/')
			{
				if(a == ' ') continue;
				temp[i] = a;
				i++;
			}
			temp[i] = 0;
			default_note = atoi(temp) * LENGTH_SCALE;
			i = 0;
			while((a = ifile.get()) != '\n')
			{
				if(a == ' ') continue;
				temp[i] = a;
				i++;
			}
			temp[i] = 0;
			default_note /= atoi(temp);
			if(verbose)
				std::cout << "Default note length set to: " << float(default_note) / float(LENGTH_SCALE) << " by L header entry.\n";
		}
		else if(a == 'Q')
		{
			std::cerr << "Setting speed not implemented yet (Q header entry).\n";
			while(ifile.get() != '\n');
		}
		else if(a == '%')	// Ignore comments.
		{
			while(ifile.get() != '\n');
		}
		else if(a == '\n')	// Ignore empty lines.
		{
		}
		else
		{
			std::cerr << "Header " << a << " ignored.\n";
			while(ifile.get() != '\n');
		}
	}
 
	ofile << "#include <stdint.h>\n" << "\n" << "const uint16_t notes[] = {";

	erase();
	
	while((a = ifile.get()) != EOF)
	{
		switch(a)
		{
		case '^':
			translate();
			note.local_modifier = +1;
			note.local_modifier_enabled = true;
			break;
		case '_':
			translate();
			note.local_modifier = -1;
			note.local_modifier_enabled = true;
			break;
		case '=':
			translate();
			note.local_modifier = 0;
			note.local_modifier_enabled = true;
			break;
		case '.':
			translate();
			note.staccato = true;
			break;
		case 'z':
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'g':
			translate();
			note.oktave = 3;
			note.base = toupper(a);
			break;
		case 'Z':
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'G':
			translate();
			note.oktave = 2;
			note.base = a;
			break;
		case ',':
			note.oktave--;
			break;
		case '\'':
			note.oktave++;
			break;
		case '-':
			note.local_legato = true;
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			if(!note.denom_current)
				note.num[note.counter++] = a;
			else
				note.denom[note.counter++] = a;
			break;
		case '/':
			if(!note.denom_current)
			{
				note.denom_current = true;
				note.counter = 0;
				note.denom[note.counter] = 2;
			}
			else if(note.denom[note.counter] == 2)
			{
				note.counter = 0;
				note.denom[note.counter] = 4;
			}
			else if(note.denom[note.counter] == 4)
			{
				note.counter = 0;
				note.denom[note.counter] = 8;
			}
			else
				std::cerr << "Failed to decode note length encoded by //...\n";
			break;
		case '%':
			while(ifile.get() != '\n');	// Ignore comments.
			break;
		case '|':
			translate();
			for(int i = 0; i < 4; i++)
				for(int j = 0; j < 7; j++)
					local_modifier_enabled[i][j] = false;
			break;
		case ' ':
		case '\t':
		case '\n':
			break;	// Ignore whitespace.
		case '[':
			std::cerr << "Encountered unimplemented inline [...] in music. Ignored.\n";
			while(ifile.get() != ']');
			break;
		case '(':
			translate();
			legato++;
			break;
		case ')':
			legato--;
			if(legato < 0)
			{
				std::cerr << "Unmatched slur ending encountered.\n";
			}
			translate();
			break;
		case ':':
		default:
			if(verbose)
				std::cerr << "Encountered unimplemented " << a << " in music.\n";
			break;
		}
	}
	translate();
	ofile << " 0xffff };\n";
	if(verbose)
		std::cout << "Wrote " << length << " notes, length in full notes: " << tlength << "\n";
	ofile.close();
	ifile.close();
	return(0);
}
